﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Cursed.geom;
using Cursed.gob;
using Cursed.gob.mob;
using Cursed.gob.tile;
using Microsoft.Xna.Framework;
using Box2D.XNA;
using Cursed.gob.mob.kob;

namespace Cursed.col
{
    public class B2DCollisionInfo
    {
        public Body body;
        public HashSet<Contact> contacts;
        public B2DCollisionInfo(Body body)
        {
            this.body = body;
            this.contacts = new HashSet<Contact>();
        }
    }

    public class B2DContactListener : IContactListener
    {
        #region IContactListener Members

        public void BeginContact(Contact contact)
        {
            Gob gob1 = (Gob)contact.GetFixtureA().GetBody().GetUserData();
            Gob gob2 = (Gob)contact.GetFixtureB().GetBody().GetUserData();
            gob1.Touch(gob2);
            gob2.Touch(gob1);
            ((B2DCollisionInfo)gob1.CollisionInfo).contacts.Add(contact);
            ((B2DCollisionInfo)gob2.CollisionInfo).contacts.Add(contact);
        }

        public void EndContact(Contact contact)
        {
            Gob gob1 = (Gob)contact.GetFixtureA().GetBody().GetUserData();
            Gob gob2 = (Gob)contact.GetFixtureB().GetBody().GetUserData();
            gob1.Untouch(gob2);
            gob2.Untouch(gob1);
            ((B2DCollisionInfo)gob1.CollisionInfo).contacts.Remove(contact);
            ((B2DCollisionInfo)gob2.CollisionInfo).contacts.Remove(contact);
        }

        public void PostSolve(Contact contact, ref ContactImpulse impulse)
        {

        }

        public void PreSolve(Contact contact, ref Manifold oldManifold)
        {

        }

        #endregion
    }
    public class B2DContactFilter : IContactFilter
    {

        #region IContactFilter Members

        public bool RayCollide(object userData, Fixture fixture)
        {
            return true;
        }

        public bool ShouldCollide(Fixture fixtureA, Fixture fixtureB)
        {
            Gob gob1 = (Gob)fixtureA.GetBody().GetUserData();
            Gob gob2 = (Gob)fixtureB.GetBody().GetUserData();

            if (gob1 is Tile && gob2 is Tile)
            {
                return false;
            }

            if (gob1.IsEffect || gob2.IsEffect)
                return false;

            if (fixtureA.IsSensor() || fixtureB.IsSensor())
            {
                return true;
            }

            if (gob1.Transcendent || gob2.Transcendent)
                return false;

            return true;
        }

        #endregion
    }

    public class Box2DCollider : QTCollider
    {
        World world;

        public Box2DCollider(CursedGame game)
            : base(game)
        {
            world = new World(Vector2.Zero, true);
            world.ContactFilter = new B2DContactFilter();
            world.ContactListener = new B2DContactListener();
        }

        void b2dAddGob(Gob gob, bool isFixed)
        {
            if (world == null)
                return;

            IShape shape = gob.Shape;
            shape.setTransform(Vector2.Zero, 0, 1);

            FixtureDef fixDef = new FixtureDef();

            float scale = 1;
            if (gob is Mob)
                scale = ((Mob)gob).MobInfo.Scale;

            if (shape is Rect)
            {
                Rect rect = (Rect)shape;
                PolygonShape poly = new PolygonShape();
                poly.SetAsBox(rect.Width / 2, scale * rect.Height / 2, scale * rect.Center, 0);
                fixDef.shape = poly;
            }
            else if (shape is Circle)
            {
                Circle circle = (Circle)shape;
                CircleShape cshape = new CircleShape();
                cshape._radius = scale * circle.Radius;
                cshape._p = circle.Center;
                fixDef.shape = cshape;
            }
            else
            {
                //complain/impossible
                return;
            }


            BodyDef b2dBodyDef = new BodyDef();
            if (gob is Tile)
                b2dBodyDef.type = BodyType.Static;
            else if (gob.IsEffect)
                b2dBodyDef.type = BodyType.Kinematic;
            else
                b2dBodyDef.type = BodyType.Dynamic;

            Body b2dBody = world.CreateBody(b2dBodyDef);

            b2dBody.SetUserData(gob);

            b2dBody.CreateFixture(fixDef);
            fixDef.isSensor = true;
            b2dBody.CreateFixture(fixDef);


            MassData md = new MassData();
            md = new MassData();
            if (!isFixed)
            {
                md.mass = ((Mob)gob).MobInfo.Mass;
                md.i = Single.MaxValue;
            }
            else
            {
                md.mass = Single.MaxValue;
                md.i = Single.MaxValue;
            }
            b2dBody.SetMassData(ref md);

            b2dBody.Position = gob.GobInfo.Loc;

            gob.CollisionInfo = new B2DCollisionInfo(b2dBody);

        }

        #region ICollider Members

        public override void Clear()
        {
            if (mobs != null)
            {
                foreach (Mob mob in mobs)
                {
                    world.DestroyBody(((B2DCollisionInfo)mob.CollisionInfo).body);
                }
            }
            if (tiles != null)
            {
                foreach (Tile tile in tiles)
                {
                    world.DestroyBody(((B2DCollisionInfo)tile.CollisionInfo).body);
                }
            }
            base.Clear();
        }

        public override void HandleCollisions()
        {

        }

        public override void Update(GameTime gameTime)
        {
            buildQT();
            /*
            foreach (Mob mob in mobs)
            {
                mob.clearTouch();
            }
            foreach (Tile tile in tiles)
            {
                tile.clearTouch();
            }
            */
            world.ClearForces();

            float seconds = gameTime.ElapsedGameTime.Milliseconds * .001f;

            foreach (Mob mob in mobs)
            {
                Body b2dBody = ((B2DCollisionInfo)mob.CollisionInfo).body;
                b2dBody.Position = mob.GobInfo.Loc;
                b2dBody.SetLinearVelocity(mob.MobInfo.Vel);
                b2dBody.ApplyForce(mob.Force, mob.GobInfo.Loc);
                mob.UpdatePosition(seconds);
            }

            for (Contact contact = world.GetContactList(); contact != null; contact = contact.GetNext())
            {
                Gob gob1 = (Gob)contact.GetFixtureA().GetBody().GetUserData();
                Gob gob2 = (Gob)contact.GetFixtureB().GetBody().GetUserData();
                B2DCollisionInfo cinfo1 = (B2DCollisionInfo)gob1.CollisionInfo;
                B2DCollisionInfo cinfo2 = (B2DCollisionInfo)gob2.CollisionInfo;
                cinfo1.contacts.Add(contact);
                cinfo2.contacts.Add(contact);
            }

            world.Step(seconds, 5, 5);

            foreach (Mob mob in mobs)
            {
                Body b2dBody = ((B2DCollisionInfo)mob.CollisionInfo).body;
                mob.GobInfo.Loc = b2dBody.Position;
                mob.MobInfo.Vel = b2dBody.GetLinearVelocity();
                mob.MobInfo.Theta = b2dBody.GetAngle();
            }
        }

        public override void AddMob(Cursed.gob.mob.Mob mob)
        {
            base.AddMob(mob);
            b2dAddGob(mob, mob.MobInfo.Stuck);
        }

        public override void RemoveMob(Cursed.gob.mob.Mob mob)
        {
            Body body = ((B2DCollisionInfo)mob.CollisionInfo).body;
            world.DestroyBody(body);
            base.RemoveMob(mob);
        }

        public override void AddTile(Cursed.gob.tile.Tile tile)
        {
            base.AddTile(tile);
            b2dAddGob(tile, true);
        }

        public override void RemoveTile(Cursed.gob.tile.Tile tile)
        {
            Body body = ((B2DCollisionInfo)tile.CollisionInfo).body;
            world.DestroyBody(body);
            base.RemoveTile(tile);
        }

        #endregion
    }
}
