﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Cursed.view;
using Cursed.input;
using Microsoft.Xna.Framework;
using Cursed.gob;
using Cursed.geom;
using Cursed.view.renders;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Cursed.gob.mob;
using Cursed.gob.mob.kob;
using Cursed.gob.tile;
using System.Reflection;

namespace Cursed.editor
{
    class CursorInfo : MobInfo
    {
        public CursorInfo()
            : base()
        {
            Depth = -50; //on top
        }
    }
    class Cursor : Mob
    {
        public Cursor(CursedGame game)
            : base(game, new CursorInfo())
        {
            shape = new Rect(Vector2.Zero, .9f, .9f);
            renderInfo = new SpriteRenderInfo(shape.Bounds);
            renderInfo.Mask = new Color(.6f, .6f, .2f, .7f);
            renderInfo.Texture = CursedGame.textureLib.GetTexture("square");
            addRender(CursedGame.getRender(Cursed.view.renders.Render.RenderType.SpriteRender), renderInfo);

            Transient = true;
            Transcendent = true;
        }

        public Vector2 CursorLoc
        {
            get { return GobInfo.Loc; }
            set
            {
                value.X = (int)value.X;
                value.Y = (int)value.Y;

                GobInfo.Loc = value;
            }
        }
    }

    class Editor : IFocus, IControllable
    {
        CursedGame game;
        ICamera camera;
        Cursor cursor;
        Vector2 loc;
        MouseController mouseController;
        MouseTrailController mouseTrailController;
        KeyDirectionController directionController;
        EscControl escapeControl;
        TextInputController textController;
        Cursed.view.Console console;

        Vector2 dragStart;
        float zoomStart;

        bool active;

        GobInfo currentInfo;
        Type destroyType = typeof(Wall);

        #region editor

        public Editor(CursedGame game, Vector2 loc)
        {
            this.game = game;
            camera = game.Camera;

            cursor = new Cursor(game);
            this.loc = loc;

            mouseController = new MouseController(camera, game.Window);
            game.AddController(mouseController);
            mouseController.Push(this);

            mouseTrailController = new MouseTrailController(camera, game.Window);
            game.AddController(mouseTrailController);
            mouseTrailController.Push(this);

            directionController = new KeyDirectionController();
            game.AddController(directionController);
            directionController.Push(this);

            escapeControl = new EscControl();
            game.AddController(escapeControl);
            escapeControl.Push(this);

            textController = new TextInputController();
            game.AddController(textController);
            textController.Push(this);

            active = false;
            console = new Cursed.view.Console(game);
            currentInfo = null;
        }

        public void activate()
        {
            game.AddGob(cursor);
            camera.PushFocus(this);
            game.Paused = true;
            active = true;
            textController.CanBeActive = true;
        }

        public void deactivate()
        {
            game.RemoveGob(cursor);
            camera.PopFocus(this);
            game.Paused = false;
            active = false;
            textController.CanBeActive = false;
        }
        
        #endregion

        #region IControllable Members

        public void control(ControlInfo info)
        {
            if (!active)
            {
                if (info is EscapeControlInfo)
                {
                    if (active)
                        deactivate();
                    else
                        activate();
                }
                return;
            }
            if (!textController.Active)
            {
                if (info is MouseCursorInfo)
                {
                    MouseCursorInfo mouseCursorInfo = (MouseCursorInfo)info;
                    cursor.CursorLoc = mouseCursorInfo.Mouse;
                    if (mouseCursorInfo.Clicked && mouseCursorInfo.Type == MouseType.Offensive)
                    {
                        create();
                    }
                    if (mouseCursorInfo.Clicked && mouseCursorInfo.Type == MouseType.Defensive)
                    {
                        destroy();
                    }
                }
                if (info is MouseTrailInfo)
                {
                    MouseTrailInfo mouseTrailInfo = (MouseTrailInfo)info;
                    if (mouseTrailInfo.Trail.Count == 1)
                    {
                        if (mouseTrailInfo.Modifier)
                        {
                            zoomStart = camera.Zoom;
                        }
                        else
                        {
                            dragStart = loc;
                        }
                    }
                    if (mouseTrailInfo.Trail.Count >= 2)
                    {
                        Vector2 first = mouseTrailInfo.Trail[0];
                        Vector2 last = mouseTrailInfo.Trail[mouseTrailInfo.Trail.Count - 1] - loc;
                        Vector2 offset = first - last;
                        
                        if (mouseTrailInfo.Modifier)
                        {
                            float z = zoomStart * (float)Math.Exp(.05 * offset.Y);
                            if (z < .05f) z = .05f;
                            if (z > 2) z = 2;
                            camera.Zoom = z;
                        }
                        else
                        {
                            loc = offset;
                        }
                    }
                }
                if (info is DirectionControlInfo)
                {
                    DirectionControlInfo directionControlInfo = (DirectionControlInfo)info;
                    loc += directionControlInfo.Direction * .05f;
                }
                if (info is EscapeControlInfo)
                {
                    if (active)
                        deactivate();
                    else
                        activate();
                }
            }
            if (info is TextInputInfo)
            {
                console.control(info);
                string message = ((TextInputInfo)info).TextInput;
                if (message != null && message.Length != 0)
                {
                    string[] commands = message.Split(';');
                    foreach (string command in commands)
                    {
                        if (!parseInput(command.Trim()))
                        {
                            report("Bad input \"" + command + "\"");
                        }
                    }
                }
            }
        }



        #endregion

        #region IFocus Members

        public Vector2 Loc
        {
            get { return loc; }
        }

        #endregion

        #region manip

        void create()
        {
            if (currentInfo != null)
            {
                GobInfo newInfo = currentInfo.Clone();
                if (newInfo is MobInfo)
                {
                    MobInfo mInfo = (MobInfo)newInfo;
                    mInfo.Loc = cursor.CursorLoc;
                }
                if (newInfo is TileInfo)
                {
                    TileInfo tInfo = (TileInfo)newInfo;
                    tInfo.X = (int)cursor.CursorLoc.X;
                    tInfo.Y = (int)cursor.CursorLoc.Y;
                }
                Gob gob = newInfo.Create(game);
                game.AddGob(gob);
                System.Console.WriteLine("added a " + gob.GetType().Name);
                destroyType = gob.GetType();
            }
        }

        void destroy()
        {
            if (destroyType == null)
                return;
            System.Console.WriteLine(cursor.Bounds);
            foreach (Gob gob in game.GobsInBounds(cursor.Bounds))
            {
                if (gob.GetType().IsSubclassOf(destroyType) || gob.GetType() == destroyType)
                {
                    if (gob is Tile)
                    {
                        Tile t = (Tile)gob;
                        System.Console.WriteLine(t.X + " " + t.Y+" - "+t.Bounds);
                    }
                    game.RemoveGob(gob);
                    System.Console.WriteLine("removed a " + gob.GetType().Name);
                }
            }
        }
        #endregion
        void report(string txt)
        {
            textController.addLine("Error: " + txt);
        }
        public void Draw(SpriteBatch batch)
        {
            console.Draw(batch);
        }
        GobInfo getPresetInfo(string infoType)
        {
            if (infoType == "tree")
            {
                ObstructionInfo info = new ObstructionInfo();
                info.Texture = "Tree1";
                info.DrawCenter = new Vector2(0, 0.5f);
                info.DrawWidth = 2;
                info.DrawHeight = 2;
                return info;
            }
            if (infoType == "boss")
            {
                DemonInfo dmi = new DemonInfo();
                dmi.MaxHealth = 5;
                dmi.CurrentHealth = 5;
                dmi.Scale = 2;
                dmi.Thrust = 8;
                dmi.StrengthFactor = 2;
                dmi.ViewRange = 12;
                dmi.LungeDuration = 1;
                dmi.LungeFactor = 4;
                return dmi;
            }
            if (infoType == "assassin")
            {

                DemonInfo dmi = new DemonInfo();
                dmi.Transparency = .2f;
                dmi.Scale = .5f;
                return dmi;
            }
            if (infoType == "walltop")
            {
                WallInfo info = new WallInfo(0, 0);
                info.Mask = new Color(.2f, .2f, .2f, 1f);
                info.Texture = "square";
                return info;
            }
            return null;
        }

        /*
         * commands:
         * .=<type> 
         * .<name>=<value>
         * s <id>
         * l <id>
         */
        bool parseInput(string input)
        {
            if (input.Length == 0)
                return true;

            #region setting-values
            if (input[0] == '.')
            {
                if (input.Length == 1)
                {
                    return false;
                }
                if (input[1] == '=')
                {
                    string typeName = input.Substring(2);
                    currentInfo = getPresetInfo(typeName);
                    if (currentInfo == null)
                        currentInfo = GobInfo.Instantiate(typeName);
                    if (currentInfo == null)
                        report("bad typename '" + typeName + "'");
                    //else
                      //  report("Type set to '" + typeName + "'");
                }
                else
                {
                    if (currentInfo == null)
                        return false;
                    string[] bits = input.Substring(1).Split('=');
                    if (bits.Length != 2)
                    {
                        return false;
                    }

                    string field = bits[0].Trim();
                    string value = bits[1].Trim();

                    if (field.Length == 0 || value.Length == 0)
                        return false;

                    currentInfo.SetField(field, value);
                }
            }
            #endregion

            #region saving

            if (input[0] == 'l')
            {
                string[] args = input.Split(' ');
                if (args.Length != 2)
                {
                    return false;
                }
                string path = args[1];
                game.LoadLevel(path);
            }

            if (input[0] == 's')
            {
                string[] args = input.Split(' ');
                if (args.Length != 2)
                {
                    return false;
                }
                string path = args[1];
                game.SaveLevel(path);
            }

            #endregion

            return true;
        }
    }
}
