﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Cursed.view;
using Cursed.input;
using Cursed.gob.tile;
using Cursed.geom;
using System.Runtime.Serialization;
using Cursed.view.renders;
using Cursed.gob.mob.kob;


namespace Cursed.gob.mob
{
    public class MobInfo : GobInfo
    {
        float mass; //for collisions, perhaps accels. applying a force would consider mass.
        bool stuck;

        Vector2 vel;
        float thrust;

        float drag;
        float maxSpeedSquared;

        float theta; //rotation
        float omega; //angular velocity

        float scale;

        float lungeDuration;
        float lungeFactor;


        public MobInfo()
            : base()
        {
            mass = 1;
            stuck = false;
            vel = Vector2.Zero;
            thrust = 1;
            drag = 0;
            maxSpeedSquared = 100;
            theta = 0;
            omega = 0;
            scale = .9f;
            lungeDuration = 0;
            lungeFactor = 3;
        }

        public float Mass
        {
            get { return mass; }
            set { mass = value; }
        }
        public bool Stuck
        {
            get { return stuck; }
            set { stuck = value; }
        }
        public Vector2 Vel
        {
            get { return vel; }
            set { vel = value; }
        }
        public float Thrust
        {
            get { return thrust; }
            set { thrust = value; }
        }
        public float Drag
        {
            get { return drag;}
            set { drag = value;}
        }
        public float MaxSpeedSquared
        {
            get { return maxSpeedSquared; }
            set { maxSpeedSquared = value; }
        }
        public float Theta
        {
            get { return theta; }
            set { theta = value; }
        }

        public float Omega
        {
            get { return omega; }
            set { omega = value; }
        }
        public float Scale
        {
            get { return scale; }
            set { scale = value; }
        }
        public float LungeDuration
        {
            get { return lungeDuration; }
            set { lungeDuration = value; }
        }
        public float LungeFactor
        {
            get { return lungeFactor; }
            set { lungeFactor = value; }
        }
       
    }


    public class Mob : Gob, IControllable, IFocus
    {

        //direction the control is telling this mob to move at the moment
        protected Vector2 direction;
        //this time step's force vector
        protected Vector2 force;
        //this time step's drag factor
        protected float drag;

        float lungeLeft;

        protected RenderInfo renderInfo;

        public Mob(CursedGame game, MobInfo info)
            : base(game, info)
        {
            lungeLeft = 0;
            force = Vector2.Zero;
            drag = 0;
            direction = Vector2.Zero;
            if (this.renderInfo != null)
            {
                this.renderInfo.Mask = info.Mask;
            }
        }


        public override void Update(GameTime gameTime)
        {
            float seconds = gameTime.ElapsedGameTime.Milliseconds*.001f;
            LungeLeft -= seconds;

            if (renderInfo != null)
            {
                updateRenderTransform(renderInfo);
            }

            //System.Console.WriteLine(GetType().Name + " " + GobInfo.Loc);

            updateTransform();

            float friction = 0;

            foreach (Gob touchedGob in touchedGobs)
            {
                if (touchedGob is Ground)
                {
                    Ground ground = (Ground)touchedGob;
                    if (ground.GroundInfo.Drag > drag)
                        drag = ground.GroundInfo.Drag;
                    if (ground.GroundInfo.Friction > friction)
                        friction = ground.GroundInfo.Friction;
                }
            }
            float thrust = MobInfo.Thrust;
            if (LungeLeft > 0)
                thrust *= MobInfo.LungeFactor;
            ApplyForce(direction * thrust * friction);

            Vector2 dragForce = -1 * drag * MobInfo.Vel;
            float maxDrag = MobInfo.Vel.Length() * MobInfo.Mass / seconds;
            if (dragForce.Length() > maxDrag)
            {
                dragForce.Normalize();
                dragForce *= maxDrag;
            }

            ApplyForce(dragForce);

            base.Update(gameTime);
        }
        public MobInfo MobInfo
        {
            get { return (MobInfo)info; }
        }

        public void ApplyForce(Vector2 f)
        {
            force += f;
        }

        public Vector2 Force
        {
            get { return force; }
        }

        public void ApplyDrag(float drag)
        {
            this.drag += drag;
        }

        public void doPhysicsCollision(Gob gob)
        {
            Vector2 refVelocity = Vector2.Zero;
            Vector2 relVelocity = MobInfo.Vel;
            if (gob is Mob)
            {
                refVelocity = ((Mob)gob).MobInfo.Vel;
            }
            relVelocity -= refVelocity;

            Vector2 normal = Shape.intersectionNormal(gob.Shape);
            Vector2 unit = normal;
            unit.Normalize();

            float relInlineSpeed = Vector2.Dot(unit, relVelocity);

            if (relInlineSpeed < 0)
            {
                Vector2 inVel = unit * relInlineSpeed;
                Vector2 offVel = relVelocity - inVel;
                
                MobInfo.Vel = refVelocity + offVel;


                float m1 = 0;
                float m2 = 1;
                if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
                {
                    Mob mob = (Mob)gob;
                    m1 = MobInfo.Mass;
                    m2 = mob.MobInfo.Mass;
                }

                MobInfo.Vel += inVel * (m1 - Const.RestitutionCoeff * m2) / (m1 + m2);

                if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
                {
                    Mob mob = (Mob)gob;
                    mob.MobInfo.Vel += inVel * (Const.RestitutionCoeff + 1) * m1 / (m1 + m2);
                }
            }

            //CursedGame.Profiler.Start("pushOff");
            //pushOff(gob, normal, 1);
            //CursedGame.Profiler.Stop("pushOff");
        }

        public void Push(float factor)
        {

            foreach (Gob touchedGob in touchedGobs)
            {
                if (touches(touchedGob) && !MobInfo.Stuck && !Transcendent && !touchedGob.Transcendent)
                {
                    Vector2 normal = Shape.intersectionNormal(touchedGob.Shape);
                    pushOff(touchedGob, normal, factor);
                    updateTransform();
                    if (touchedGob is Mob)
                    {
                        ((Mob)touchedGob).updateTransform();
                    }
                }
            }
        }

        /*
         * Push self off of gob in small increments, until no longer touching
         */
        protected void pushOff(Gob gob, Vector2 normal, float factor)
        {

            if (normal == Vector2.Zero)
                return;// normal = new Vector2((float)CursedGame.Random.NextDouble(), (float)CursedGame.Random.NextDouble());

            Vector2 relV = MobInfo.Vel;
            if (gob is Mob)
                relV -= ((Mob)gob).MobInfo.Vel;
            if (Vector2.Dot(relV, normal) > 0)
                return;

            Vector2 offset = GobInfo.Loc - gob.GobInfo.Loc;
            if (Vector2.Dot(offset, normal) < 0)
                return;

            Vector2 pushDistance = normal * factor;
            if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
            {
                pushDistance *= .5f;
                gob.GobInfo.Loc -= pushDistance;
            }
            GobInfo.Loc += pushDistance;
            
        }

        protected void updateTransform()
        {
            if (shape != null)
                shape.setTransform(MobInfo.Loc, MobInfo.Theta, MobInfo.Scale);
        }

        protected void updateRenderTransform(RenderInfo rInfo)
        {
            rInfo.Transform = Microsoft.Xna.Framework.Matrix.CreateScale(MobInfo.Scale);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateRotationZ(MobInfo.Theta);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateTranslation(MobInfo.Loc3);
        }

        public void UpdatePosition(float seconds)
        {


            foreach (Gob touchedGob in touchedGobs)
            {
                if (!MobInfo.Stuck && !Transcendent && !touchedGob.Transcendent)
                {
                    doPhysicsCollision(touchedGob);
                }
            }

            Vector2 accelaration = force / MobInfo.Mass;

            //v = v0 + at
            MobInfo.Vel += accelaration * seconds;

            //pos = p0 + v*t + 1/2 a*t^2;
            MobInfo.Loc += MobInfo.Vel * seconds + accelaration * .5f * seconds * seconds;

            MobInfo.Theta += MobInfo.Omega * seconds;

            force = Vector2.Zero;
            drag = MobInfo.Drag;
        }

        protected virtual void performLunge()
        {
            if (LungeLeft < -MobInfo.LungeDuration * 3)
            {
                LungeLeft = MobInfo.LungeDuration;
            }
        }

        public virtual void control(ControlInfo cInfo)
        {
            if (cInfo is DirectionControlInfo)
            {
                direction = ((DirectionControlInfo)cInfo).Direction;
            }
            if (cInfo is LungeInfo)
            {
                performLunge();
            }
        }

        public float LungeLeft
        {
            get { return lungeLeft; }
            set { lungeLeft = value; }
        }

        public bool Lunging
        {
            get { return lungeLeft > 0; }
        }
        
        #region IFocus Members

        public Vector2 Loc
        {
            get { return MobInfo.Loc; }
        }

        #endregion
    }
}
