﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Cursed.input.ai;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Cursed.geom;
using Cursed.input;
using Cursed.view.renders;
using Cursed.sound;

namespace Cursed.gob.mob.kob
{
    public class DemonInfo : KobInfo
    {
        int tailLength;
        float decay;
        bool spinAttack;
        float strength;

        public DemonInfo()
            : base()
        {
            ViewRange = 5;
            Thrust = 5;
            Drag = 2f;
            tailLength = 10;
            Transparency = 1f;
            decay = .1f;
            Energy = -1f;
            Mass = Scale;
            Depth = Const.DemonDepth;
            spinAttack = false;
            strength = 1;
        }

        public override Gob Create(CursedGame game)
        {
            return new Demon(game, this);
        }

        public float StrengthFactor
        {
            get { return strength; }
            set { strength = value; }
        }
        public int TailLength
        {
            get { return tailLength; }
            set { tailLength = value; }

        }
        public float Decay
        {
            get { return decay;}
            set { decay = value; }
        }
        public bool SpinAttack
        {
            get { return spinAttack; }
            set { spinAttack = value; }
        }
    }

    public class Demon : Kob
    {  
        protected TrailRenderInfo trailInfo;     

        protected Sound basicAttackSound;
        protected Sound spinAttackSound;

        public TrailRenderInfo TrailInfo
        {
            get { return trailInfo; }
            set { trailInfo = value; }
        }
        public Demon(CursedGame game, DemonInfo info)
            : base(game, info)
        {
            basicAttackSound = null;
            spinAttackSound = null;
            Transient = false;
            Transcendent = false;
        }
        public override void Load()
        {
            base.Load();
            shape = new Circle(Vector2.Zero, .5f);
            renderInfo = new SpriteRenderInfo(new Bounds(GobInfo.DrawCenter, GobInfo.DrawWidth, GobInfo.DrawHeight));
            renderInfo.Mask = Microsoft.Xna.Framework.Graphics.Color.Gray;
            renderInfo.Texture = CursedGame.textureLib.GetTexture("circle");

            trailInfo = new TrailRenderInfo();

            addRender(CursedGame.getRender(Cursed.view.renders.Render.RenderType.TrailRender), trailInfo);
            addRender(CursedGame.getRender(Cursed.view.renders.Render.RenderType.SpriteRender), renderInfo);

            hitSound = SoundIDs.DemonHit(CursedGame);
            basicAttackSound = SoundIDs.FamiliarBasicAttack(CursedGame);
            deathSound = SoundIDs.DemonDeath(CursedGame);

            addControllers();
        }
        protected virtual void addControllers()
        {
            DruidSeeker seekerControl = new DruidSeeker(this);
            CursedGame.AddController(seekerControl);
            seekerControl.Push(this);

            DruidAttacker attackerControl = new DruidAttacker(this);
            CursedGame.AddController(attackerControl);
            attackerControl.Push(this);
        }
        /*
        public override bool touches(Gob other)
        {
            if(!lunge || other.Equals(goal))
                return base.touches(other);
            return false;
        }
         */
        protected virtual void performBasicAttack(Vector2 direction)
        {
            if (Ready)
            {
                basicAttackSound.stop();
                basicAttackSound.play();
                Vector2 loc = GobInfo.Loc + direction * (.75f);

                float theta = (float)Math.Acos(direction.X);
                if (direction.Y < 0) theta *= -1;
                theta -= MathHelper.PiOver2;

                AttackInfo ati = new AttackInfo();
                ati.Loc = loc;
                ati.Vel = MobInfo.Vel;
                ati.Theta = theta;
                ati.Energy = KobInfo.Energy;
                ati.Strength = Const.DemonClawStrength*DemonInfo.StrengthFactor;
                ati.Duration = Const.DemonClawAttackDuration;
                Circle c = new Circle(Vector2.Zero, .5f);
                Attack at = new ClawsAttack(CursedGame, ati, this, c);
                CursedGame.AddGob(at);
                CoolDown += Const.DemonClawAttackCoolDown;
            }
        }
        protected void performSpinAttack()
        {
            float theta = (float)Math.Acos(direction.X);
                if (direction.Y < 0) theta *= -1;
                theta -= MathHelper.PiOver2;
            if (Ready)
            {
                spinAttackSound.stop();
                spinAttackSound.play();
                AttackInfo ati = new AttackInfo();
                ati.Loc = GobInfo.Loc;
                ati.Vel = MobInfo.Vel;
                ati.Theta = theta;
                ati.Energy = KobInfo.Energy;
                ati.Strength = Const.DemonSpinStrength * DemonInfo.StrengthFactor;
                ati.Duration = Const.DemonSpinAttackDuration;
                Circle c = new Circle(Vector2.Zero, this.KobInfo.Scale * .75f);
                Attack at = new SpinAttack(CursedGame, ati, this, c);
                CursedGame.AddGob(at);
                CoolDown += Const.DemonSpinAttackCoolDown;

            }
        }
        public override void control(Cursed.input.ControlInfo cInfo)
        {
 	        base.control(cInfo);
            if (cInfo is BasicAttackInfo)
            {
                if(((BasicAttackInfo)cInfo).Type == BasicAttackInfo.ATTACKTYPE.NORMAL)
                    performBasicAttack(((BasicAttackInfo)cInfo).Direction);
                else if (((BasicAttackInfo)cInfo).Type == BasicAttackInfo.ATTACKTYPE.SPIN)
                {
                    performSpinAttack();
                }
            }
        }
        public override void Update(GameTime gameTime)
        {
            if (DemonInfo.Energy > -1f)
            {
                DemonInfo.Energy -= gameTime.ElapsedGameTime.Milliseconds * .001f * DemonInfo.Decay;
                if (DemonInfo.Energy > 0)
                {
                    DemonInfo.Energy = 0f;
                }
            }

            HashSet<Mob> moreSeenMobs = new HashSet<Mob>();
            foreach (Mob mob in SeenMobs)
            {
                if (mob is Familiar)
                {
                    Familiar familiar = (Familiar)mob;
                    if (familiar.Master != null)
                        moreSeenMobs.Add(familiar.Master);
                }
            }
            SeenMobs.UnionWith(moreSeenMobs);

            if (Lunging)
            {
                foreach (Gob gob in TouchedGobs)
                {
                    if (gob is Kob)
                    {
                        Vector2 diff = gob.GobInfo.Loc - GobInfo.Loc;
                        diff.Normalize();
                        performBasicAttack(diff);
                    }
                }
            }

            updateColor();
            trailInfo.Update((SpriteRenderInfo)renderInfo);

            base.Update(gameTime);
        }
        public virtual void updateColor()
        {
            float val = (1 + DemonInfo.Energy);
            renderInfo.Mask = new Color((1 - val) * .3f, (1 - val) * .3f + val*.6f, (1 - val) * .3f, DemonInfo.Transparency); 
        }
        public DemonInfo DemonInfo
        {
            get { return (DemonInfo)info; }
        }

        public override void Unload()
        {
           
            base.Unload();
        }
    }
}
