﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Cursed.gob.tile;
using Cursed.gob;
using Cursed.gob.mob;
using Cursed.geom;

namespace Cursed.input.ai
{
    public class PathControl : Controller
    {
        struct TCoord
        {
            public int x;
            public int y;
            public TCoord(int x, int y)
            {
                this.x = x;
                this.y = y;
            }
            public TCoord(Vector2 loc)
            {
                x = (int)Math.Floor(loc.X + .5f);
                y = (int)Math.Floor(loc.Y + .5f);
            }
            public Vector2 Vector2
            {
                get { return new Vector2(x, y); }
                set
                {
                    x = (int)Math.Floor(value.X + .5f);
                    y = (int)Math.Floor(value.Y + .5f);
                }
            }
            public bool Equals(TCoord second)
            {
                return x == second.x && y == second.y;
            }
        }

        CursedGame game;

        float epsilon;
        bool active;
        Vector2 destination;
        int depth;
        Dictionary<TCoord, float> scoreTable;
        int forgetThreshold;
        int ticks;

        public PathControl(CursedGame game)
        {
            this.game = game;
            epsilon = .1f;
            active = false;
            destination = Vector2.Zero;
            depth = 15;
            forgetThreshold = 100;
            ticks = 0;
            scoreTable = new Dictionary<TCoord, float>();
        }

        public bool Active
        {
            get { return active; }
            set { active = value; }
        }

        public Vector2 Destination
        {
            get { return destination; }
            set { destination = value; }
        }

        TCoord DestinationTC
        {
            get { return new TCoord(destination); }
        }

        public int Depth
        {
            get { return depth; }
            set { depth = value; }
        }

        public int ForgetThreshold
        {
            get { return forgetThreshold; }
            set { forgetThreshold = value; }
        }

        protected void forget()
        {
            scoreTable.Clear();
        }

        bool isPassable(TCoord tc)
        {
            foreach (Tile tile in game.TilesAt(tc.x, tc.y))
            {
                if (!tile.Transcendent && !(tile is Door))
                    return false;
            }
            return true;
        }

        float obstruction(TCoord tc)
        {
            foreach (Tile tile in game.TilesAt(tc.x, tc.y))
            {
                if (!tile.Transcendent && !(tile is Door))
                    return Single.PositiveInfinity;
            }
            float obs = 0;
            foreach (Mob mob in game.MobsInBounds(new Bounds(tc.Vector2, 1, 1)))
            {
                if (mob.Transcendent || mob == game.DruidPlayer)
                    continue;
                if (mob.MobInfo.Stuck)
                    obs += 100;
                obs += 5f;
            }
            return obs;
        }

        float getScore(TCoord tc)
        {
            if (tc.Equals(DestinationTC))
                return 0;
            if (scoreTable.ContainsKey(tc))
                return scoreTable[tc];
            float obs = 0;
            if (!DestinationTC.Equals(tc))
                obs += obstruction(tc);
            return scoreTable[tc] = (tc.Vector2 - Destination).Length() + obs;
        }

        TCoord ARTDPStep(TCoord tc)
        {
            if (tc.Equals(DestinationTC))
                return tc;

            float bestScore = Single.PositiveInfinity;
            TCoord bestTC = tc;

            for (int x = tc.x - 1; x <= tc.x + 1; x++)
            {
                for (int y = tc.y - 1; y <= tc.y + 1; y++)
                {
                    TCoord n = new TCoord(x, y);
                    float ns = getScore(n);
                    if (ns < bestScore)
                    {
                        bestScore = ns;
                        bestTC = n;
                    }
                }
            }

            if (getScore(tc) != Single.PositiveInfinity)
            {
                if (bestTC.Equals(tc))
                    scoreTable[tc] = Single.PositiveInfinity;
                else
                    scoreTable[tc] = bestScore + (tc.Vector2 - bestTC.Vector2).Length();
            }

            TCoord destTC = bestTC;
            if (game.Random.NextDouble() < epsilon)
            {
                destTC.x = game.Random.Next(-1, 2);
                destTC.y = game.Random.Next(-1, 2);
            }
            if (destTC.Equals(tc))
                destTC = bestTC;

            return destTC;
        }

        protected void runTrajectory(Vector2 start)
        {
            TCoord curTC = new TCoord(start);

            int stepsLeft = Depth;
            while (stepsLeft != 0 && !curTC.Equals(DestinationTC))
            {
                curTC = ARTDPStep(curTC);
                stepsLeft--;
            }
        }

        protected Vector2 getDirection(Vector2 start)
        {
            ticks++;
            if (ticks % forgetThreshold == 0)
                forget();
            game.Profiler.Start("PathControl.getDirection");
            //do one run of ARTDP for each act called, stopping on destination or depth threshold
            runTrajectory(start);

            //check what the next step for the current situation is
            TCoord shortTermTC = ARTDPStep(new TCoord(start));
            Vector2 diff = shortTermTC.Vector2 - start;
            if (shortTermTC.Equals(DestinationTC))
                diff = Destination - start;
            if (diff != Vector2.Zero)
                diff.Normalize();
            game.Profiler.Stop("PathControl.getDirection");
            return diff;
        }

        public override void act(Microsoft.Xna.Framework.GameTime time)
        {
            if (!active)
                return;
            if (Target == null || !(Target is Gob))
                return;
            Gob targetGob = (Gob)Target;

            DirectionControlInfo di = new DirectionControlInfo();
            di.Direction = getDirection(targetGob.GobInfo.Loc);
            Target.control(di);
        }

        public override void cleanup()
        {

        }
    }
}
